# Copyright of the Board of Trustees of Columbia University in the City of New York
'''
Prepares field map data (from a Siemens scanner) for phase unwrapping with FSL
\nAuthor: Marina Manso Jimeno
\nLast updated: 07/08/2020
'''
import os
import scipy.io as sio
import numpy as np
import nibabel as nib

from pydicom import dcmread

from OCTOPUS.Recon.rawdata_recon import mask_by_threshold

def phase_unwrap_prep(data_path_raw, data_path_dicom, dst_folder, dTE):
    '''
    Prepares a Siemens fieldmap for phase unwrapping using FSL. Saves a phase difference image and a magnitude image with
    the ROI extracted in niftii format to use as inputs for FSL.

    Parameters
    ----------
    data_path_raw : str
        Path to the MATLAB/npy file containing the field map raw data with dimensions [lines, columns, echoes, channels]
    data_path_dicom : str
        Path to the DICOM file containing the phase difference image (.IMA)
    dst_folder : str
        Path to the destination folder where the niftii files are saved
    dTE : float
        Difference in TE between the two echoes in seconds
    '''
    if data_path_raw[-3:] == 'mat':
        b0_map = sio.loadmat(data_path_raw)['b0_map'] # field map raw data
    elif data_path_raw[-3:] == 'npy':
        b0_map = np.load(data_path_raw)
    else:
        raise ValueError('File format not supported, please input a .mat or .npy file')

    ##
    # Acq parameters
    ##
    N = b0_map.shape[1] # Matrix Size
    Nchannels = b0_map.shape[-1]

    if len(b0_map.shape) < 5:
        Nslices = 1
        b0_map = b0_map.reshape(b0_map.shape[0], N, 1, 2, Nchannels)
    else:

        Nslices = b0_map.shape[-2]

    ##
    # FT to get the echo complex images
    ##

    echo1 = np.zeros((N * 2, N, Nslices, Nchannels), dtype=complex)
    echo2 = np.zeros((N * 2, N, Nslices, Nchannels), dtype=complex)
    for ch in range(Nchannels):
        for sl in range(Nslices):
            echo1[:, :, sl, ch] = np.fft.fftshift(np.fft.ifft2(b0_map[:, :, sl, 0, ch]))
            echo2[:, :, sl, ch] = np.fft.fftshift(np.fft.ifft2(b0_map[:, :, sl, 1, ch]))

    # Crop the lines from oversampling factor of 2
    oversamp_factor = int(b0_map.shape[0] / 4)
    echo1 = echo1[oversamp_factor:-oversamp_factor, :, :, :]
    echo2 = echo2[oversamp_factor:-oversamp_factor, :, :, :]

    # Magnitude image with object 'brain' extracted
    mag_im = np.sum(np.abs(echo1), 3)
    mag_im = np.divide(mag_im, np.max(mag_im))
    brain_mask = mask_by_threshold(mag_im)
    brain_extracted = np.squeeze(mag_im) * brain_mask
    img = nib.Nifti1Image(brain_extracted, np.eye(4))
    nib.save(img, os.path.join(dst_folder, 'mag_vol_extracted.nii.gz'))

    # Phase difference image from DICOM data
    data = dcmread(data_path_dicom)
    # Get the image data
    vol = data.pixel_array
    # Save as niftii
    img = nib.Nifti1Image(np.rot90(vol), np.eye(4))
    nib.save(img, os.path.join(dst_folder, 'phase_diff.nii.gz'))



# Copyright of the Board of Trustees of Columbia University in the City of New York
'''
Methods to read DICOM files corresponding to GRE field map and GRE Siemens sequences
\nAuthor: Marina Manso Jimeno
\nLast modified: 07/16/2020
'''

from pydicom import dcmread
import numpy as np

def read_dicom(path):
    '''
    Reads dicom file from path

    Parameters
    ----------
    path : str
        Path of the file

    Returns
    -------
    vol : np.ndarray
        Array containing the image volume
    '''
    data = dcmread(path)
    # Get the image data
    vol = data.pixel_array
    vol = np.fliplr(vol)
    return vol





# Copyright of the Board of Trustees of Columbia University in the City of New York
'''
Methods to do k-space to image and image to k-space for both Cartesian and non-cartesian data
\nAuthor: Marina Manso Jimeno
\nLast updated: 07/18/2020
'''

import numpy.fft as npfft
import numpy as np
import pynufft

from math import pi

def im2ksp(M, cartesian_opt, NufftObj=None, params=None):
    '''Image to k-space transformation

    Parameters
    ----------
    M : numpy.ndarray
        Image data
    cartesian_opt : int
        Cartesian = 1, Non-Cartesian = 0.
    NufftObj : pynufft.linalg.nufft_cpu.NUFFT_cpu
        Non-uniform FFT Object for non-cartesian transformation. Default is None.
    params : dict
        Sequence parameters. Default is None.

    Returns
    -------
    kspace : numpy.ndarray
        k-space data
    '''

    if cartesian_opt == 1:

        kspace = npfft.fftshift(npfft.fft2(M))
    elif cartesian_opt == 0:
        # Sample phantom along ktraj
        if 'Npoints' not in params:
            raise ValueError('The number of acquisition points is missing')
        if 'Nshots' not in params:
            raise ValueError('The number of shots is missing')
        kspace = NufftObj.forward(M).reshape((params['Npoints'], params['Nshots']))  # sampled image
    else:
        raise ValueError('Cartesian option should be either 0 or 1')
    return kspace

def ksp2im(ksp, cartesian_opt, NufftObj=None, params=None):
    '''K-space to image transformation

    Parameters
    ----------
    ksp : numpy.ndarray
        K-space data
    cartesian_opt : int
        Cartesian = 1, Non-Cartesian = 0.
    NufftObj : pynufft.linalg.nufft_cpu.NUFFT_cpu
        Non-uniform FFT Object for non-cartesian transformation. Default is None.
    params : dict
        Sequence parameters. Default is None.

    Returns
    -------
    im : numpy.ndarray
        Image data
    '''
    if cartesian_opt == 1:
        im = npfft.ifft2(npfft.fftshift(ksp))

    elif cartesian_opt == 0:
        if 'dcf' in params:
            ksp_dcf = ksp.reshape((params['Npoints']*params['Nshots'],))*params['dcf']
            im = NufftObj.adjoint(ksp_dcf)  # * np.prod(sqrt(4 * params['N'] ** 2))
        else:
            ksp_dcf = ksp.reshape((params['Npoints'] * params['Nshots'],))
            im = NufftObj.solve(ksp_dcf, solver='cg', maxiter=50)



    else:
        raise ValueError('Cartesian option should be either 0 or 1')

    return im

def nufft_init(kt, params):
    '''Initializes the Non-uniform FFT object

    Parameters
    ----------
    kt : numpy.ndarray
        K-space trajectory
    params : dict
        Sequence parameters.

    Returns
    -------
    NufftObj : pynufft.linalg.nufft_cpu.NUFFT_cpu
        Non-uniform FFT Object for non-cartesian transformation
    '''
    kt_sc = pi / abs(np.max(kt))
    kt = kt * kt_sc # pyNUFFT scaling [-pi, pi]
    om = np.zeros((params['Npoints'] * params['Nshots'], 2))
    om[:, 0] = np.real(kt).flatten()
    om[:, 1] = np.imag(kt).flatten()

    NufftObj = pynufft.NUFFT_cpu()  # Create a pynufft object
    Nd = (params['N'], params['N'])  # image size
    Kd = (2 * params['N'], 2 * params['N'])  # k-space size
    Jd = (6, 6)  # interpolation size

    NufftObj.plan(om, Nd, Kd, Jd)
    return NufftObj


# Copyright of the Board of Trustees of Columbia University in the City of New York
'''
Methods to simulate different types of field maps
\nAuthor: Marina Manso Jimeno
\nLast modified: 07/16/2020
'''

import numpy as np
import cv2
import math
import matplotlib.pyplot as plt

import OCTOPUS.ORC as ORC

def fieldmap_bin(field_map, bin):
    '''
    Bins a given field map given a binning value

    Parameters
    ----------
    field_map : numpy.ndarray
        Field map matrix in Hz
    bin : int
        Binning value in Hz

    Returns
    -------
    binned_field_map : numpy.ndarray
        Binned field map matrix
    '''
    fmax = field_map.max()
    bins = np.arange(-fmax, fmax + bin, bin)
    binned_field_map = np.zeros(field_map.shape)
    for x in range(field_map.shape[0]):
        for y in range(field_map.shape[1]):
            idx = ORC.find_nearest(bins, field_map[x, y])
            binned_field_map[x, y] = bins[idx]

    return binned_field_map

def parabola_formula(N):
    """
    Parabola values to fit an image of N rows/columns

    Parameters
    ----------
    N :  int
        Matrix size

    Returns
    -------
    yaxis : numpy.ndarray
        y axis values of the parabola
    """
    x1, y1 = -N / 10, 0.5
    x2, y2 = 0, 0
    x3, y3 = N / 10, 0.5

    denom = (x1 - x2) * (x1 - x3) * (x2 - x3)
    A = (x3 * (y2 - y1) + x2 * (y1 - y3) + x1 * (y3 - y2)) / denom
    B = (x3 * x3 * (y1 - y2) + x2 * x2 * (y3 - y1) + x1 * x1 * (y2 - y3)) / denom
    C = (x2 * x3 * (x2 - x3) * y1 + x3 * x1 * (x3 - x1) * y2 + x1 * x2 * (x1 - x2) * y3) / denom

    y = lambda x: A * x ** 2 + B * x + C
    yaxis = y(np.arange(-N / 2, N / 2, 1)).reshape(1, N)

    return yaxis

def parabolic(N, fmax, bin_opt = True, bin_val = 5):
    """
    Creates a parabolic field map

    Parameters
    ----------
    N : int
        Field map dimensions (NxN)
    fmax : float
        Frequency range in Hz
    bin_opt : bool
        Binning option. Default is True
    bin_val : int
        Binning value. Default is 5 Hz

    Returns
    -------
    field map : numpy.ndarray
        Field map matrix with dimensions [NxN] and scaled from -fmax to +fmax Hz
    """
    y = parabola_formula(N)
    rows = np.tile(y, (N,1))
    field_map_mat = rows + rows.T
    dst = np.zeros(field_map_mat.shape)
    field_map = cv2.normalize(field_map_mat, dst, -fmax, fmax, cv2.NORM_MINMAX)
    if bin_opt:
        field_map = fieldmap_bin(field_map, bin_val)

    return field_map

def hyperbolic(N, fmax, bin_opt = True,  bin_val= 5):
    """
    Creates a hyperbolic field map

    Parameters
    ----------
    N : int
       Field map dimensions (NxN)
    fmax : float
       Frequency range in Hz
    bin_opt : bool
        Binning option. Default is True
    bin_val : int
        Binning value, default is 5 Hz


    Returns
    -------
    field map : numpy.ndarray
       Field map matrix with dimensions [NxN] and scaled from -fmax to +fmax Hz
    """
    y = parabola_formula(N)
    rows = np.tile(y, (N, 1))
    field_map_mat = rows - rows.T
    dst = np.zeros(field_map_mat.shape)
    field_map = cv2.normalize(field_map_mat, dst, -fmax, fmax, cv2.NORM_MINMAX)
    if bin_opt:
        field_map = fieldmap_bin(field_map, bin_val)
    return field_map

def realistic(im , fmax , bin_opt = True, bin_val = 5):
    """
    Creates a realistic field map based on the input image

    Parameters
    ----------
    im : numpy.ndarray
        Input image
    fmax : float
       Frequency range in Hz
    bin_opt : bool
        Binning option. Default is True
    bin_val : int
        Binning value, default is 5 Hz

    Returns
    -------
    field_map : numpy.ndarray
        Field map matrix with dimensions same as im and scaled from -fmax to +fmax Hz
    """
    if im.shape[0] != im.shape[1]:
        raise ValueError('Images have to be squared (NxN)')

    N = im.shape[0]
    hist = np.histogram(im)
    mask1= cv2.threshold(im, hist[1][hist[0].argmax()+1], 1, cv2.THRESH_BINARY)
    mask2 = cv2.threshold(im, hist[1][hist[0].argmax()], 1, cv2.THRESH_BINARY_INV)
    mask = mask1[1] + mask2[1]

    M = np.random.rand(2, 2)
    M2 = cv2.resize(M, (N, N))


    dst = np.zeros(M2.shape)
    field_map = cv2.normalize(M2, dst, -fmax, fmax, cv2.NORM_MINMAX) * mask

    if bin_opt:
        field_map = fieldmap_bin(field_map, bin_val)

    return field_map

def spherical_order4(N, fmax, bin_opt  = True,  bin_val= 5):
    """
    Creates a field map simulating spherical harmonics of 4th order

    Parameters
    ----------
    N : int
       Field map dimensions (NxN)
    fmax : float
       Frequency range in Hz
    bin_opt : bool
        Binning option. Default is True
    bin_val : int
        Binning value, default is 5 Hz

    Returns
    -------
    field_map : numpy.ndarray
        Field map matrix with dimensions NxN and scaled from -fmax to +fmax Hz
    """
    offset = 1 #math.sqrt(1 / (4 * math.pi))
    x = np.tile(np.linspace(-offset, offset, N).reshape(N, 1), (1, N))
    l1_X = np.dstack([x] * N)
    #plot_views(l1_X, 'X')

    l1_Z = np.ones((N, N, N)) * np.linspace(-offset, offset, N).reshape(1, 1, N)
    #plot_views(l1_Z, 'Z')

    y = np.flipud(x.T)
    l1_Y = np.dstack([y] * N)
    #plot_views(l1_Y, 'Y')

    middle_slice = int(round(N / 2)) - 1
    l4_X4 = 105 * (l1_X ** 2 - l1_Y ** 2) ** 2 - 420 * l1_X ** 2 * l1_Y **2
    #plot_views(l4_X4, 'X4')
    field_map_mat = np.squeeze(l4_X4[:,:,middle_slice])

    dst = np.zeros(field_map_mat.shape)
    field_map = cv2.normalize(field_map_mat, dst, -fmax, fmax, cv2.NORM_MINMAX)
    if bin_opt:
        field_map = fieldmap_bin(field_map, bin_val)
    return field_map


'''def plot_views(matrix, fig_title):
    
    middle_slice = int(round(matrix.shape[0] / 2)) - 1
    fig, axes = plt.subplots(nrows=1, ncols=3)
    fig.suptitle(fig_title, fontsize=16)
    # plt.subplot(131)
    axes[0].imshow(np.rot90(matrix[:, :, middle_slice]), vmin=matrix.min(), vmax=matrix.max())
    axes[0].set_title('Axial (XY)')
    # plt.subplot(132)
    axes[1].imshow(np.rot90(matrix[:, middle_slice, :]), vmin=matrix.min(), vmax=matrix.max())
    axes[1].set_title('Coronal (XZ)')
    # plt.subplot(133)
    axes[2].imshow(np.rot90(matrix[middle_slice, :, :]), vmin=matrix.min(), vmax=matrix.max())
    axes[2].set_title('Sagittal (YZ)')

    # plt.colorbar()
    plt.show()'''


# Copyright of the Board of Trustees of Columbia University in the City of New York
'''
Author: Marina Manso Jimeno
Last modified: 07/21/2020
'''
import numpy as np
import scipy.io as sio
import nibabel as nib

def get_data_from_file(input):
    '''
    Get the data of a file given a string or a dictionary independently of the data format

    Parameters
    ----------
    input : str or dict
        Input path or dictionary in case of user upload (Colab)

    Returns
    -------
    file_data : np.ndarray
        Data extracted from the file
    '''
    if isinstance(input, str):
        file_name = input
    elif isinstance(input, dict):
        file_name = next(iter(input))
    file_format = file_name[file_name.find('.'):]
    if file_format == '.mat':
        file_data_dict = sio.loadmat(file_name)
        file_data = file_data_dict[list(file_data_dict.keys())[-1]]
    elif file_format == '.npy':
        file_data = np.load(file_name)
    elif file_format == '.nii.gz':
        file_data = nib.load(file_name).get_fdata()
    else:
        raise ValueError('Sorry, this file format is not supported:' + (file_format))

    return file_data


# Copyright of the Board of Trustees of Columbia University in the City of New York
'''
Methods to reconstruct a field map and spiral images from raw data.
\nAuthor: Marina Manso Jimeno
\nLast modified: 07/16/2020
'''
import scipy.io as sio
import numpy as np
import matplotlib.pyplot as plt
import numpy.fft as fft
import math

from OCTOPUS.utils.get_data_from_file import get_data_from_file
from OCTOPUS.Recon.imtransforms import nufft_init, ksp2im

def mask_by_threshold(im):
    '''
    Masks a magnitude image by thresholding according to:
    Jenkinson M. (2003). Fast, automated, N-dimensional phase-unwrapping algorithm. Magnetic resonance in medicine, 49(1), 193–197. https://doi.org/10.1002/mrm.10354

    Parameters
    ----------
    im : np.ndarray
        Magnitude image

    Returns
    -------
    mask : np.ndarray
        Binary image for background segmentation
    '''
    im = np.squeeze(im)
    mask = np.ones(im.shape)
    threshold = 0.1 * (np.percentile(im, 98) - np.percentile(im, 2)) + np.percentile(im, 2)
    mask[np.where(im < threshold)] = 0
    return mask

def hermitian_product(echo1, echo2, dTE):
    '''
    Calculates the phase difference and frequency map given two echo data and deltaTE between them.
    Channels are combined using the hermitian product as described by:
    Robinson, S., & Jovicich, J. (2011). B0 mapping with multi-channel RF coils at high field. Magnetic resonance in medicine, 66(4), 976–988. https://doi.org/10.1002/mrm.22879

    Parameters
    ----------
    echo1 : np.ndarray
        Complex image corresponding to the first echo with dimensions [N, N, Nslices, Nchannels]
    echo2 : np.ndarray
        Complex image corresponding to the second echo with same dimensions as echo1
    dTE : float
        TE difference between the two echos in seconds.

    Returns
    -------
    fmap : np.ndarray
        Frequency map in Hz with dimensions [N, N, Nslices]
    '''

    delta_theta = np.angle(np.sum(echo2 * np.conjugate(echo1), axis=-1))
    fmap = - delta_theta / (dTE * 2 * math.pi)

    return fmap

def separate_channels(echo1, echo2, dTE):
    '''
    Calculates the phase difference and frequency map given two echo data and deltaTE between them.
    Channels are combined using a trimmed average method inspired by the separate channels method described by:
    Robinson, S., & Jovicich, J. (2011). B0 mapping with multi-channel RF coils at high field. Magnetic resonance in medicine, 66(4), 976–988. https://doi.org/10.1002/mrm.22879

    Parameters
    ----------
    echo1 : np.ndarray
        Complex image corresponding to the first echo with dimensions [N, N, Nslices, Nchannels]
    echo2 : np.ndarray
        Complex image corresponding to the second echo with same dimensions as echo1
    dTE : float
        TE difference between the two echos in seconds.

    Returns
    -------
    fmap_chcomb : np.ndarray
        Frequency map in Hz with dimensions [N, N, Nslices]
    '''
    echo1_ph = np.angle(echo1)
    echo2_ph = np.angle(echo2)

    delta_theta = echo2_ph - echo1_ph
    fmap = - delta_theta / (2 * math.pi * dTE)
    fmap_chcomb = np.zeros(fmap.shape[:-1])
    for sl in range(fmap.shape[-2]):
        for i in range(fmap.shape[0]):
            for j in range(fmap.shape[1]):
                voxel_vals = fmap[i,j,sl,:]
                lowest_quartile = np.quantile(voxel_vals, 0.25)
                highest_quartile = np.quantile(voxel_vals, 0.75)
                trimmed_vals = [val for val in voxel_vals if val > lowest_quartile and val < highest_quartile]
                fmap_chcomb[i,j,sl] = np.mean(np.asarray(trimmed_vals))

    return fmap_chcomb

def fmap_recon(data, dTE, method = 'HP', save = 0, plot = 0, dst_folder = None):
    '''
    Frequency map reconstruction from dual echo raw data

    Parameters
    ----------
    data : str or np.ndarray
        Path containing the field map file or array containing the field map data
    dTE : float
        Difference in TE between the two echoes in seconds
    method : str
        Method for channel combination. Options are 'HP' or 'SC'. Default is 'HP'.
    plot : bool, Optional
        Plotting a slice of the reconstructed frequency map option. Default is 0 (not plot).
    save : bool, Optional
        Saving the data in a .npy file option. Default is 0 (not save).
    dst_folder : str, Optional
        Path to the folder where the reconstructed field map is saved. Default is None.

    Returns
    -------
    fmap : np.ndarray
        Frequency map in Hz
    '''
    ##
    # Load the raw data
    ##
    if isinstance(data, str):
        f_map = get_data_from_file(data)
    elif isinstance(data, np.ndarray):
        f_map = data

    ##
    # Acq parameters
    ##
    N = f_map.shape[1] # Matrix Size
    Nchannels = f_map.shape[-1]

    if len(f_map.shape) < 5:
        Nslices = 1
        f_map = f_map.reshape(f_map.shape[0], N, 1, 2, Nchannels)
    else:

        Nslices = f_map.shape[2]

    ##
    # FT to get the echo complex images
    ##

    echo1 = np.zeros((N * 2, N, Nslices, Nchannels), dtype=complex)
    echo2 = np.zeros((N * 2, N, Nslices, Nchannels), dtype=complex)
    for ch in range(Nchannels):
        for sl in range(Nslices):
            echo1[:, :, sl, ch] = fft.fftshift(fft.ifft2(f_map[:, :, sl, 0, ch]))
            echo2[:, :, sl, ch] = fft.fftshift(fft.ifft2(f_map[:, :, sl, 1, ch]))

    # Crop the lines from oversampling factor of 2
    oversamp_factor = int(f_map.shape[0] / 4)
    echo1 = echo1[oversamp_factor:-oversamp_factor, :, :, :]
    echo2 = echo2[oversamp_factor:-oversamp_factor, :, :, :]

    ##
    # Calculate the field map
    ##
    if method == 'HP':
        fmap = hermitian_product(echo1, echo2, dTE)
    elif method == 'SC':
        fmap = separate_channels(echo1, echo2, dTE)
    else:
        raise ValueError('The method you specified is not supported')

    fmap = np.fliplr(fmap)
    if save:
        if dst_folder is None:
            raise ValueError('Please specify destination folder')
        np.save(dst_folder + 'fmap', fmap)

    if plot:
        mid = math.floor(fmap.shape[-1]/2)
        plt.imshow(np.rot90(fmap[:, :, mid], -1), cmap='gray')
        #plt.imshow(np.rot90(np.fliplr(b0map[:,:,0]), -1), cmap='gray') # rotate only for display purposes
        plt.axis('off')
        plt.title('Field Map')
        plt.colorbar()
        plt.show()

    return fmap

def spiral_recon(data, ktraj, N, plot = 0, save = 0, dst_folder = None):
    '''
    Spiral image reconstruction from raw data

    Parameters
    ----------
    data : str or np.ndarray
        Path containing the raw data file of array containing the raw data
    ktraj : np.ndarray
        k-space trajectory coordinates with dimensions [Npoints, Nshots]
    N : int
        Matrix size of the reconstructed image
    plot : bool, Optional
        Plotting a slice of the reconstructed image option. Default is 0 (not plot).
    save : bool, Optional
        Saving the data in a .npy file option. Default is 0 (not save).
    dst_folder : str, Optional
        Path to the folder where the reconstructed image is saved. Default is None.
    '''

    ##
    # Load the raw data
    ##
    if isinstance(data, str):
        dat = get_data_from_file(data)
    elif isinstance(data, np.ndarray):
        dat = data
    ##
    # Acq parameters
    ##
    Npoints = ktraj.shape[0]
    Nshots = ktraj.shape[1]
    Nchannels = dat.shape[-1]

    if len(dat.shape) < 4:
        Nslices = 1
        dat = dat.reshape(Npoints, Nshots, 1, Nchannels)
    else:
        Nslices = dat.shape[-2]


    if dat.shape[0] != ktraj.shape[0] or dat.shape[1] != ktraj.shape[1]:
        raise ValueError('Raw data and k-space trajectory do not match!')


    ##
    # Recon
    ##
    NUFFT_object = nufft_init(ktraj, {'N': N, 'Npoints': Npoints, 'Nshots': Nshots})
    im = np.zeros((N, N, Nslices, Nchannels), dtype=complex)
    for ch in range(Nchannels):
        for sl in range(Nslices):
            im[:,:,sl,ch] = ksp2im(dat[:,:,sl,ch], 0, NUFFT_object, {'N': N, 'Npoints': Npoints, 'Nshots': Nshots})#NufftObj.solve(dat[:,:,sl,ch].flatten(), solver='cg', maxiter=50)


    sos = np.sum(np.abs(im), -1)
    sos = np.divide(sos, np.max(sos))

    if plot:
        plt.imshow(np.rot90(np.abs(sos[:,:,0]),-1), cmap='gray')
        plt.axis('off')
        plt.title('Uncorrected Image')
        plt.show()

    if save:
        if dst_folder is None:
            raise ValueError('Please specify destination folder')
        np.save(dst_folder + 'uncorrected_spiral.npy', sos)
    return sos


# Copyright of the Board of Trustees of Columbia University in the City of New York
'''
Methods for ORC (Off-Resonance Correction) and off-resonance simulation
\nAuthor: Marina Manso Jimeno
\nLast updated: 07/14/2020
'''

import numpy.fft as npfft
import numpy as np

from math import ceil, pi

from OCTOPUS.Recon.imtransforms import ksp2im, im2ksp, nufft_init

def add_or(M, kt, df, nonCart = None, params = None):
    '''Forward model for off-resonance simulation

    Parameters
    ----------
    M : numpy.ndarray
        Image data
    kt : numpy.ndarray
        k-space trajectory
    df : numpy.ndarray
        Field map
    nonCart : int , optional
        Cartesian/Non-Cartesian trajectory option. Default is None.
    params : dict , optional
        Sequence parameters. Default is None.

    Returns
    -------
    M_or : numpy.ndarray
        Off-resonance corrupted image data
    '''

    '''Create a phase matrix - 2*pi*df*t for every df and every t'''
    if nonCart is not None:
        cartesian_opt = 0
        NufftObj = nufft_init(kt, params)

    else:
        cartesian_opt = 1
        NufftObj = None
        params = None

    kspace = im2ksp(M, cartesian_opt, NufftObj, params)
    M_or = np.zeros(M.shape, dtype=complex)

    for x in range(M.shape[0]):
        for y in range(M.shape[1]):
            phi = 2*pi*df[x,y]*kt
            kspace_orc = kspace*np.exp(-1j*phi)
            M_corr = ksp2im(kspace_orc, cartesian_opt, NufftObj, params)
            M_or[x,y] = M_corr[x,y]

    return M_or

def add_or_CPR(M, kt, df, nonCart = None, params = None):
    '''Forward model for off-resonance simulation. The number of fourier transforms = number of unique values in the field map.

    Parameters
    ----------
    M : numpy.ndarray
        Image data
    kt : numpy.ndarray
        k-space trajectory
    df : numpy.ndarray
        Field map
    nonCart : int , optional
        Cartesian/Non-Cartesian trajectory option. Default is None.
    params : dict , optional
        Sequence parameters. Default is None.

    Returns
    -------
    M_or : numpy.ndarray
        Off-resonance corrupted image data
    '''
    # Create a phase matrix - 2*pi*df*t for every df and every t
    if nonCart is not None:
        cartesian_opt = 0
        NufftObj = nufft_init(kt, params)
        T = np.tile(params['t_vector'], (1, kt.shape[1]))

    else:
        cartesian_opt = 1
        NufftObj = None
        params = None
        T = kt

    kspace = im2ksp(M, cartesian_opt, NufftObj, params)

    df_values = np.unique(df)

    M_or_CPR = np.zeros((M.shape[0], M.shape[1], len(df_values)), dtype=complex)
    #kspsave= np.zeros((params['Npoints'],params['Nshots'],len(df_values)),dtype=complex)
    for i in range(len(df_values)):
        phi = - 2 * pi* df_values[i] * T
        kspace_or= kspace * np.exp(1j * phi)
        #kspsave[:,:,i] = kspace_or
        M_corr = ksp2im(kspace_or, cartesian_opt, NufftObj, params)
        M_or_CPR[:, :, i] = M_corr

    M_or = np.zeros(M.shape, dtype=complex)
    for x in range(M.shape[0]):
        for y in range(M.shape[1]):
            fieldmap_val = df[x, y]
            idx = np.where(df_values == fieldmap_val)
            M_or[x, y] = M_or_CPR[x, y, idx]
    '''plt.imshow(abs(M_or))
    plt.show()'''
    return M_or#, kspsave

def orc(M, kt, df):
    '''Off-resonance correction for Cartesian trajectories

    Parameters
    ----------
    M : numpy.ndarray
        Cartesian k-space data
    kt : numpy.ndarray
        Cartesian k-space trajectory
    df : numpy.ndarray
        Field map

    Returns
    -------
    M_hat : numpy.ndarray
        Off-resonance corrected image data
    '''

    kspace = npfft.fftshift(npfft.fft2(M))
    M_hat = np.zeros(M.shape, dtype=complex)
    for x in range(M.shape[0]):
        for y in range(M.shape[1]):
            phi = 2 * pi * df[x, y] * kt
            kspace_orc = kspace * np.exp(1j * phi)
            M_corr = npfft.ifft2(kspace_orc)
            M_hat[x, y] = M_corr[x, y]

    return M_hat

def CPR(dataIn, dataInType, kt, df, nonCart=None, params=None):
    '''Off-resonance Correction by Conjugate Phase Reconstruction

    Parameters
    ----------
    dataIn : numpy.ndarray
        k-space raw data or image data
    dataInType : str
        Can be either 'raw' or 'im'
    kt : numpy.ndarray
        k-space trajectory.
    df : numpy.ndarray
        Field map
    nonCart : int
        Non-cartesian trajectory option. Default is None (Cartesian).
    params : dict
        Sequence parameters. Default is None (Cartesian).

    Returns
    -------
    M_hat : numpy.ndarray
        Corrected image data.
    '''

    if nonCart is not None:
        cartesian_opt = 0
        NufftObj = nufft_init(kt, params)
        T = np.tile(params['t_vector'], (1, kt.shape[1]))
        N = params['N']
    else:
        cartesian_opt = 1
        NufftObj = None
        T = kt
        N = dataIn.shape[0]

    if dataInType == 'im':
        if dataIn.shape[0] != dataIn.shape[1]:
            raise ValueError('Data dimensions do not agree with expected image dimensions')
        rawData = im2ksp(dataIn, cartesian_opt, NufftObj, params)
    elif dataInType == 'raw':
        if dataIn.shape[0] != kt.shape[0] or dataIn.shape[1] != kt.shape[1]:
            raise ValueError('Data dimensions do not agree with expected dimensions (same as ktraj)')
        rawData = dataIn
    else:
        raise ValueError('The type of input data should be either raw or im')

    df_values = np.unique(df)
    M_CPR = np.zeros((N, N, len(df_values)), dtype=complex)
    for i in range(len(df_values)):
        phi = 2 * pi * df_values[i] * T
        kspace_orc = rawData * np.exp(1j * phi)
        M_corr = ksp2im(kspace_orc, cartesian_opt, NufftObj, params)
        M_CPR[:, :, i] = M_corr

    M_hat = np.zeros((N, N), dtype=complex)
    for x in range(df.shape[0]):
        for y in range(df.shape[1]):
            fieldmap_val = df[x,y]
            idx = np.where(df_values == fieldmap_val)
            M_hat[x,y] = M_CPR[x,y,idx]

    return M_hat

def fs_CPR(dataIn, dataInType, kt, df, Lx, nonCart= None, params= None):
    '''Off-resonance Correction by frequency-segmented Conjugate Phase Reconstruction

    Parameters
    ----------
    dataIn : numpy.ndarray
        k-space raw data or image data
    dataInType : str
        Can be either 'raw' or 'im'
    kt : numpy.ndarray
        k-space trajectory
    df : numpy.ndarray
        Field map
    Lx : int
        L (frequency bins) factor
    nonCart : int
        Non-cartesian trajectory option. Default is None (Cartesian).
    params : dict
        Sequence parameters. Default is None (Cartesian).

    Returns
    -------
    M_hat : numpy.ndarray
        Corrected image data.
    '''
    if nonCart is not None:
        cartesian_opt = 0
        NufftObj = nufft_init(kt, params)
        T = np.tile(params['t_vector'], (1, kt.shape[1]))
        N = params['N']
        t_ro = T[-1, 0] - T[0, 0]  # T[end] - TE
    else:
        cartesian_opt = 1
        NufftObj = None
        T = kt
        t_ro = T[0, -1] - T[0, 0]
        N = dataIn.shape[0]

    if dataInType == 'im':
        if dataIn.shape[0] != dataIn.shape[1]:
            raise ValueError('Data dimensions do not agree with expected image dimensions')
        rawData = im2ksp(dataIn, cartesian_opt, NufftObj, params)
    elif dataInType == 'raw':
        if dataIn.shape[0] != kt.shape[0] or  dataIn.shape[1] != kt.shape[1]:
            raise ValueError('Data dimensions do not agree with expected dimensions (same as ktraj)')
        rawData = dataIn
    else:
        raise ValueError('The type of input data should be either raw or im')

    # Number of frequency segments
    df_max = max(np.abs([df.max(), df.min()])) # Hz
    L = ceil(4 * df_max * 2 * pi * t_ro / pi)
    L= L * Lx
    f_L = np.linspace(df.min(), df.max(), L + 1) # Hz

    # T = np.tile(params['t_vector'], (1, kt.shape[1]))

    # reconstruct the L basic images
    M_fsCPR = np.zeros((N, N, L + 1),dtype=complex)
    for l in range(L+1):
        phi = 2 * pi * f_L[l] * T
        kspace_L = rawData * np.exp(1j * phi)
        M_fsCPR[:,:,l] = ksp2im(kspace_L, cartesian_opt, NufftObj, params)

    # final image reconstruction
    M_hat = np.zeros((N, N), dtype=complex)
    for i in range(M_hat.shape[0]):
        for j in range(M_hat.shape[1]):
            fieldmap_val = df[i, j]
            closest_fL_idx = find_nearest(f_L, fieldmap_val)

            if fieldmap_val == f_L[closest_fL_idx]:
                pixel_val = M_fsCPR[i, j, closest_fL_idx]
            else:
                if fieldmap_val < f_L[closest_fL_idx]:
                    f_vals = [f_L[closest_fL_idx - 1], f_L[closest_fL_idx]]
                    im_vals = [M_fsCPR[i, j, closest_fL_idx - 1], M_fsCPR[i, j, closest_fL_idx]]
                else:
                    f_vals = [f_L[closest_fL_idx], f_L[closest_fL_idx + 1]]
                    im_vals = [M_fsCPR[i, j, closest_fL_idx], M_fsCPR[i, j, closest_fL_idx + 1]]

                pixel_val = np.interp(fieldmap_val, f_vals, im_vals)

            M_hat[i, j] = pixel_val

    return M_hat

def MFI(dataIn, dataInType, kt , df, Lx , nonCart= None, params= None):
    '''Off-resonance Correction by Multi-Frequency Interpolation

    Parameters
    ----------
    dataIn : numpy.ndarray
        k-space raw data or image data
    dataInType : str
        Can be either 'raw' or 'im'
    kt : numpy.ndarray
        k-space trajectory
    df : numpy.ndarray
        Field map
    Lx : int
        L (frequency bins) factor
    nonCart : int
        Non-cartesian trajectory option. Default is None (Cartesian).
    params : dict
        Sequence parameters. Default is None.

    Returns
    -------
    M_hat : numpy.ndarray
        Corrected image data.
    '''
    if nonCart is not None:
        cartesian_opt = 0
        NufftObj = nufft_init(kt, params)
        t_vector = params['t_vector']
        T = np.tile(params['t_vector'], (1, kt.shape[1]))
        t_ro = T[-1,0] - T[0,0] # T[end] - TE
        N = params['N']
    else:
        cartesian_opt = 1
        NufftObj = None
        t_vector = kt[0].reshape(kt.shape[1],1)
        T = kt
        t_ro = T[0, -1] - T[0,0]
        N = dataIn.shape[0]

    if dataInType == 'im':
        if dataIn.shape[0] != dataIn.shape[1]:
            raise ValueError('Data dimensions do not agree with expected image dimensions')
        rawData = im2ksp(dataIn, cartesian_opt, NufftObj, params)
    elif dataInType == 'raw':
        if dataIn.shape[0] != kt.shape[0] or dataIn.shape[1] != kt.shape[1]:
            raise ValueError('Data dimensions do not agree with expected dimensions (same as ktraj)')
        rawData = dataIn
    else:
        raise ValueError('The type of input data should be either raw or im')

    df = np.round(df, 1)
    idx, idy = np.where(df == -0.0)
    df[idx, idy] = 0.0

    # Number of frequency segments
    df_max = max(np.abs([df.max(), df.min()]))  # Hz
    df_range = (df.min(), df.max())
    L = ceil(df_max * 2 * pi * t_ro / pi)
    L = L * Lx
    f_L = np.linspace(df.min(), df.max(), L + 1)  # Hz

    #T = np.tile(params['t_vector'], (1, kt.shape[1]))

    # reconstruct the L basic images
    M_MFI = np.zeros((N, N, L + 1), dtype=complex)
    for l in range(L + 1):
        phi = 2 * pi * f_L[l] * T
        kspace_L = rawData * np.exp(1j * phi)
        M_MFI[:, :, l] = ksp2im(kspace_L, cartesian_opt, NufftObj, params)

    # calculate MFI coefficients
    coeffs_LUT = coeffs_MFI_lsq(kt, f_L, df_range, t_vector)

    # final image reconstruction
    M_hat = np.zeros((N, N), dtype=complex)
    for i in range(M_hat.shape[0]):
        for j in range(M_hat.shape[1]):
            fieldmap_val = df[i,j]
            val_coeffs = coeffs_LUT[str(fieldmap_val)]
            M_hat[i, j] = sum(val_coeffs * M_MFI[i, j, :])

    return M_hat

def find_nearest(array, value):
    '''Finds the index of the value's closest array element

    Parameters
    ----------
    array : numpy.ndarray
        Array of values
    value : float
        Value for which the closest element has to be found

    Returns
    -------
    idx : int
        Index of the closest element of the array
    '''
    array = np.asarray(array)
    diff = array - value
    if value >= 0:
        idx = np.abs(diff).argmin()
    else:
        idx = np.abs(diff[array < 1]).argmin()

    return idx

def coeffs_MFI_lsq(kt, f_L, df_range, t_vector):
    '''Finds the coefficients for Multi-frequency interpolation method by least squares approximation.

    Parameters
    ----------
    kt : numpy.ndarray
        K-space trajectory
    f_L : numpy.ndarray
        Frequency segments array.
    df_range : tuple
        Frequency range of the field map (minimum and maximum).
    params : dict
        Sequence parameters.

    Returns
    -------
    cL : dict
        Coefficients look-up-table.
    '''
    fs = 0.1 # Hz
    f_sampling = np.round(np.arange(df_range[0], df_range[1]+fs, fs), 1)

    alpha = 1.2 #
    t_limit = t_vector[-1]

    T = np.linspace(0, alpha * t_limit, len(t_vector)).reshape(-1, )

    A = np.zeros((kt.shape[0], f_L.shape[0]), dtype=complex)
    for l in range(f_L.shape[0]):
        phi = 2 * pi * f_L[l] * T
        A[:, l] = np.exp(1j * phi)

    cL={}
    for fs in f_sampling:
        b = np.exp(1j * 2 * pi * fs * T)
        C = np.linalg.lstsq(A, b, rcond=None)
        if fs == -0.0:
            fs = 0.0
        cL[str(fs)] = C[0][:].reshape((f_L.shape[0]))

    return cL



# Copyright of the Board of Trustees of Columbia University in the City of New York
'''
Methods to calculate off-resonance correction performance metrics
\nAuthor: Marina Manso Jimeno
\nLast updated: 06/03/2020
'''

import scipy.io as sio
import cv2

from skimage.metrics import peak_signal_noise_ratio as pSNR
from skimage.metrics import structural_similarity as SSIM


import numpy as np
import matplotlib.pyplot as plt

from skimage import data, img_as_float

def SSIM_calc(im1, im2):
    '''
    Structural Similarity Index Calculation for two images

    Parameters
    ----------
    im1 : nupy.ndarray
        Image 1. Reference image, same shape as im2
    im2 : numpy.ndarray
        Image 2

    Returns
    -------
    hfen : float
        Measured SSIM value
    '''
    ssim = SSIM(im1, im2, data_range=im2.max() - im2.min())
    return ssim

def pSNR_calc(im1,im2):
    '''
    Peak Signal to Noise Ratio Calculation for two images

    Parameters
    ----------
    im1 : nupy.ndarray
        Image 1. Reference image, same shape as im2
    im2 : numpy.ndarray
        Image 2

    Returns
    -------
    hfen : float
        Measured SSIM value
    '''
    psnr = pSNR(im1, im2, data_range=im2.max() - im2.min())
    return psnr

def LoG(im):
    '''
    Laplacian of a Gaussian implementation with kernel size (15, 15) and sigma equal to 1.5 pixels

    Parameters
    ----------
    im : numpy.ndarray
        Input image

    Returns
    -------
    log_im : numpy.ndarray
        Laplacian of a gaussian of the image
    '''
    # rotationally symmetric kernel
    log_kernel = sio.loadmat('../utils/log_kernel.mat')['log_kernel']
    log_im = cv2.filter2D(np.abs(im), -1, log_kernel)

    return log_im

def HFEN(im1, im2):
    '''
    High Frequency Error Norm calculation for two images

    Parameters
    ----------
    im1 : nupy.ndarray
        Image 1. Reference image, same shape as im2
    im2 : numpy.ndarray
        Image 2

    Returns
    -------
    hfen : float
        Measured HFEN value
    '''
    log1 = LoG(im1)
    log2 = LoG(im2)

    hfen = np.linalg.norm((log1 - log2), 2) / np.linalg.norm(log2, 2)

    return hfen

def create_table(stack_of_images, col_names, franges):
    '''
    Displays a table with the metrics for images corrected using the different ORC methods

    Parameters
    ----------
    stack_of_images : numpy.ndarray
        stack_of_images[0] : ground truth. stack_of_images[1] : CPR corrected. stack_of_images[2] : fs-CPR corrected. stack_of_images[3] : MFI corrected.
    col_names : tuple
        Names for the columns
    franges :  tuple
        Frequency ranges of the original field map
    '''

    nmetrics = 2 #3
    nmethods = len(stack_of_images) - 1

    if len(col_names) != nmethods:
        raise ValueError('Number of methods does not match the number of columns for the table')
    if isinstance(franges, tuple):
        nfranges = len(franges)
    else:
        franges = [franges]
        nfranges = 1
        stack_of_images = np.expand_dims(stack_of_images, axis=3)

    data = np.zeros((nfranges, nmetrics, nmethods))

    for fr in range(nfranges):
        ims_fr = np.squeeze(stack_of_images[:,:,:,fr])
        GT = ims_fr[0]
        for col in range(nmethods):
            corr_im = ims_fr[col + 1]
            data[fr, 0, col] = pSNR(GT, corr_im, data_range=corr_im.max() - corr_im.min())
            data[fr, 1, col] = SSIM(GT, corr_im, data_range=corr_im.max() - corr_im.min())
            #data[fr, 2, col] = HFEN(GT, corr_im)

    # Get some pastel shades for the colors
    colors = plt.cm.BuPu(np.linspace(0.25, 0.7, nfranges))
    labels = ['pSNR', 'SSIM', 'HFEN']
    fig, ax = plt.subplots(nmetrics, 1, sharex=True)
    x = range(nmethods)
    plt.xticks(x, col_names)

    for fr in range(nfranges):
        data_fr = np.squeeze(data[fr,:,:])
        for row in range(nmetrics):
            ax[row].plot(x, data_fr[row, :], 'o-', color=colors[fr], label=franges[fr] if row == 0 else "")
            ax[row].set_ylabel(labels[row])
            if row != nmetrics -1:
                ax[row].xaxis.set_visible(False)
    fig.legend()

    if nfranges == 1:
        table_data = np.round(data[0],2)
        row_labels = labels
        row_colors = colors
    else:

        #col_names= list(col_names)
        #col_names.insert(0, 'Freq range')
        table_data = np.zeros((nfranges * nmetrics, nmethods )).tolist()
        row_count = 0
        row_labels = []
        row_colors = []
        for metric in range(nmetrics):
            for fr in range(nfranges):
                row_labels.append(franges[fr])
                row_colors.append(colors[fr])
                table_data[row_count][:] = np.round(data[fr, metric, :],2)
                row_count += 1

    the_table = plt.table(cellText = table_data,
                          cellLoc= 'center',

                          rowLabels=row_labels,
                          rowColours=row_colors,
                          colLabels=col_names,
                          loc='bottom')

    h = the_table.get_celld()[(0, 0)].get_height()
    w = the_table.get_celld()[(0, 0)].get_width() / (nmethods + 1)
    header = [the_table.add_cell(pos, -2, w, h, loc="center", facecolor="none") for pos in range(1, nfranges * nmetrics + 1)]
    count = 0
    for i in range(0, nfranges * nmetrics, nmetrics +1):
        header[i].visible_edges = "TLR"
        header[i+1].visible_edges = "LR"
        header[i+2].visible_edges = "LR"
        header[i + 1].get_text().set_text(labels[count])
        count += 1
    header[-1].visible_edges = 'BLR'

    the_table._bbox = [0, -2, 1, 1.7]
    the_table.auto_set_font_size(False)
    the_table.set_fontsize(10)
    the_table.scale(1.5, 1.5)
    plt.subplots_adjust(bottom=0.45, left=0.2)

    fig.suptitle('Performance metrics')
    plt.show()



# Copyright of the Board of Trustees of Columbia University in the City of New York
'''
Methods to plot the resulting images from off-resonance correction
\nAuthor: Marina Manso Jimeno
\nLast updated: 06/03/2020
'''
import numpy as np
import matplotlib.pyplot as plt

from mpl_toolkits.axes_grid1 import ImageGrid

def plot_correction_results(im_stack, col_names, row_names):
    '''
    Creates a plot with the resulting correction images in a grid

    Parameters
    ----------
    im_stack : numpy.ndarray
        Stack of images. [0] Corrupted images, [1]-[len(im_stack] corrected images using the different methods
    col_names : tuple
        Titles for the columns of the plot. Correction methods.
    row_names : tuple
        Titles for the rows of the plot. Off-resonance frequency ranges.
    '''
    if len(row_names) == 1:
        im_stack = np.expand_dims(im_stack, axis=-1)

    nrows = im_stack.shape[-1]
    ncols = im_stack.shape[0]

    im_list = []
    for frange in range(nrows):
        for method in range(ncols):
            im_list.append(np.abs(np.squeeze(im_stack[method, :, :, frange])))

    fig = plt.figure()
    grid = ImageGrid(fig, 111,  # similar to subplot(111)
                     nrows_ncols=(nrows, ncols),  # creates 2x2 grid of axes
                     axes_pad=0,  # pad between axes in inch.
                     )

    for ax, im, c in zip(grid, [im for im in im_list], [count for count in range(len(im_list))]):
        # Iterating over the grid returns the Axes.
        ax.imshow(im, cmap='gray')
        ax.xaxis.set_visible(False)
        ax.yaxis.set_visible(False)
        if range(ncols).count(c):
            ax.set_title(col_names[c])
        if list(np.arange(nrows)*ncols).count(c):
            ax.yaxis.set_visible(True)
            ax.yaxis.set_ticks([])
            ax.yaxis.set_label_text(row_names[int(c/ncols)], fontsize=12)

    plt.show()




import setuptools

try:  # Unicode decode error on Windows
    with open("README.md", "r") as fh:
        long_description = fh.read()
except:
    long_description = 'Off-resonance correction of MR images'

with open('requirements.txt', 'r') as f:
    install_reqs = f.read().strip()
    install_reqs = install_reqs.split("\n")

setuptools.setup(
    name = 'MR-OCTOPUS',         # How you named your package folder (MyLib)   # Chose the same as "name"
    version = '0.1.3',      # Start with a small number and increase it with every change you make
    author = 'Marina Manso Jimeno',                   # Type in your name
    author_email = 'mm5290@columbia.edu',      # Type in your E-Mail
    description = 'Off-resonance correction of MR images',   # Give a short description about your library
    long_description = long_description,
    long_description_content_type = "text/markdown",
    url = 'https://github.com/imr-framework/OCTOPUS',   # Provide either the link to your github or to your website
    packages = setuptools.find_packages(),
    include_package_data = True,
    install_requires = install_reqs,
    license = 'License :: OSI Approved :: GNU Affero General Public License v3',
    classifiers = [
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
)